@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003 - 2017 GraphicsMagick Group
+% Copyright (C) 2003 - 2018 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 % Copyright 1991-1999 E. I. du Pont de Nemours and Company
 %
@@ -574,6 +574,9 @@ static unsigned int WriteGRAYImage(const ImageInfo *image_info,Image *image)
   QuantumType
     quantum_type;
 
+  size_t
+    image_list_length;
+
   /*
     Open output image file.
   */
@@ -581,6 +584,7 @@ static unsigned int WriteGRAYImage(const ImageInfo *image_info,Image *image)
   assert(image_info->signature == MagickSignature);
   assert(image != (Image *) NULL);
   assert(image->signature == MagickSignature);
+  image_list_length=GetImageListLength(image);
   status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
   if (status == False)
     ThrowWriterException(FileOpenError,UnableToOpenFile,image);
@@ -684,7 +688,7 @@ static unsigned int WriteGRAYImage(const ImageInfo *image_info,Image *image)
     if (image->next == (Image *) NULL)
       break;
     image=SyncNextImageInList(image);
-    if (!MagickMonitorFormatted(scene++,GetImageListLength(image),&image->exception,
+    if (!MagickMonitorFormatted(scene++,image_list_length,&image->exception,
                                 SaveImagesText,image->filename))
       break;
   } while (image_info->adjoin);
